'''
This module defines the class :class:`Board`.
'''
from collections import OrderedDict
from itertools import dropwhile

from .ship import ShotResult, Ship


class Board:
    '''
    A Board is a 10x10 grid that holds :class:`Ship` s.
    '''
    def __init__(self, ships: [Ship], validate=True):
        '''
        :param [Ship] ships: the ships to place on the Board.
        :param bool validate: whether the board configuration must be valid.

        :raises ValueError: if the board configurations is invalid and `validate`
            is True.
        '''
        self._tries = OrderedDict() # type: [(int, int)] : ShotResult

        self._ships = ships
        self.ships_left = len(ships)

        if not self.valid_board and validate:
            raise ValueError('Invalid ship configuration!')

    @property
    def valid_board(self) -> bool:
        '''
        Returns a boolean indicating whether the Board's state is valid.

        A Board is valid iff it contains ten ships:

         - One Ship with length 5,
         - Two Ships with length 4,
         - Three Ships with Length 3, and
         - Four Ships with Length 2.

        Additionally, the Ships may not overlap each other, or exceed the
        boundaries of the Board.

        :type: bool
        '''
        if len(self._ships) < 10:
            return False

        fields = []
        for ship in self._ships:
            fields += ship.fields

        if len(fields) != len(set(fields)):
            return False

        correct = [2, 2, 2, 2, 3, 3, 3, 4, 4, 5]
        actual = list(map(lambda x: x.size, self._ships))

        if sorted(actual) != correct:
            return False

        return True

    @property
    def is_loss(self) -> bool:
        '''
        Whether all the Ships on the Board have sunk.

        :type: bool
        '''
        return self.ships_left <= 0

    def process_hit(self, x: int, y: int) -> ShotResult:
        '''
        Process a incoming shot by checking it against all Ships on the Board.
        If the position has already been tried, return the previous result.
        '''

        if (x, y) in self._tries:
            return self._tries[(x, y)]

        results = []
        for ship in self._ships:
            results.append(ship.process_hit(x, y))

        theresult = ShotResult.MISS

        hit_ship = list(dropwhile(lambda x: x == ShotResult.MISS, results))
        if hit_ship:
            theresult = hit_ship[0]

            if theresult == ShotResult.SUNK:
                self.ships_left -= 1

        self._tries[(x, y)] = theresult
        return theresult

    def __repr__(self) -> str:
        return self.prettyprint()

    def prettyprint(self, blind: bool = True) -> str:
        '''
        Prints the current state of the Board in a 10x10 character grid.

        :param bool blind: whether the Ships should be hidden.
        '''
        result = [10*['~'] for _ in range(10)]

        if not blind:
            for index, ship in enumerate(self._ships):
                for x, y in ship.fields:
                    result[y][x] = str(index)

        for shot, sresult in self._tries.items():
            x, y = shot
            result[y][x] = sresult.char
        return '\n'.join([''.join(row) for row in result])



'''
This module defines the class :class:`Game`.
'''
from .board import Board
from .ship import Ship, ShotResult

class Game:
    '''
    A Game represents a game of Battleships between two players.
    '''

    def __init__(self, player1: str, player2: str):
        '''
        :param str player1: the identifier to use for player one.
        :param str player2: the identifier to use for player two.
        '''

        self.turn = 0
        self.player1 = player1
        self.player2 = player2

        self.players = {
            player1: 1,
            player2: 2,
            1: player1,
            0: player2
        }

        self.boards = {
            player1: None,
            player2: None
        }

    def setup_board(self, player_id: str, ships: [Ship]) -> bool:
        '''
        Create a Board from the given set of ships, assert validity and
        register the Board with the Game if it is valid.

        Returns False if the Board is invalid, and will then not accept the
        Board.

        :param str player_id: the player identifier to set the board for.
        :param [Ship] ships: the list of Ships to populate the Board with.

        :raises AssertionError: if the passed player identifier is unknown.
        '''

        try:
            board = Board(ships)
        except ValueError:
            # Board invalid, abort
            return False

        assert player_id in self.boards

        self.boards[player_id] = board
        return True


    def register_board(self, player_id: str, board: Board):
        '''
        Register a board object to a player

        :param str player_id: the player identifier to set the board for.
        :param Board board: the board to register to the player.

        :raises AssertionError: if the passed player identifier is unknown.
        '''

        assert player_id in self.boards
        self.boards[player_id] = board


    def start_game(self) -> bool:
        '''
        Asserts both boards are present and valid, starts the Game, and gives
        the turn to player 1.

        Returns False if one or more of the Boards is absent or invalid, True
        on success.
        '''

        for _player, board in self.boards.items():
            if not (board and board.valid_board):
                return False

        self.turn = 1
        return True

    @property
    def current_player(self) -> str:
        '''
        Returns the identifier of the player on move.

        :type: str
        '''

        if self.turn == 0:
            return None

        return self.players[self.turn % 2]

    @property
    def current_adversary(self) -> str:
        '''
        Returns the identifier of the player not on move.

        :type: str
        '''

        if self.turn == 0:
            return None

        return self.players[(self.turn + 1) % 2]

    def process_fire(self, x: int, y: int) -> ShotResult:
        '''
        Process a player's input: fire at the opponent's Board, and return the
        result of the shot.

        Returns :attr:`ShotResult.LOSS` if this shot makes the opponent lose
        the game.

        :param int x: the x-coordinate to fire at.
        :param int y: the y-coordinate to fire at.
        '''

        player_id = self.current_adversary
        board = self.boards[player_id]
        result = board.process_hit(x, y)
        self.turn += 1

        if result == ShotResult.SUNK and board.is_loss:
            return ShotResult.LOSS

        return result



'''
This module defines the class :class:`Ship`, and a helper class,
:class:`ShotResult`.
'''
import enum
import re


class ShotResult(enum.IntEnum):
    ''' A ShotResult represents the result of firing on a :class:`Ship`. '''
    MISS = 0
    ''' Shot did not hit Ship. '''
    HIT = 1
    ''' Shot hit Ship, did not sink yet. '''
    SUNK = 2
    ''' Shot hit Ship, Ship sunk. '''
    LOSS = 3
    ''' Shot hit Ship, Ship sunk, Ship was last Ship on Board. '''

    @property
    def char(self) -> str:
        ''' Return the character to use when pretty-printing. '''
        return ['*', 'X', '#', 'B^u'][self]

class Ship:
    '''
    A Ship represents a Battleship piece.
    '''

    def __init__(self, x: int, y: int, horizontal: bool, size: int):
        '''
        :param int x: The x-coordinate of the top-left square.
        :param int y: The y-coordinate of the top-left square.
        :param bool horizontal: Whether the ship is positioned horizontally
            (True), or vertically (False).
        :param int size: The number of squares the Ship occupies.

        :raises ValueError: if the Ship is placed outside of the Board.
        :raises ValueError: if the Ship has an invalid `size`.
        '''
        self._x = x
        self._y = y

        self._horizontal = horizontal # type: bool
        self._size = size

        self._hit = 0

        # Check boundaries
        out_of_bounds = any(
            [
                horizontal and (x+size-1 >= 10),
                not horizontal and (y+size-1 >= 10),
                x < 0,
                y < 0
            ]
        )
        if out_of_bounds:
            raise ValueError(
                "Can't place a Ship outside of the board: {}, {}".format(
                    x+size, y)
                )

        if size <= 1 or size > 5:
            raise ValueError("Ship size outside [2..5]: {}".format(size))

    @property
    def sunk(self) -> bool:
        '''
        A ship sinks when every position is hit, which means it has been hit
        `size` times. Once the ship has sunk, this property is True.

        :type: bool
        '''
        return self._hit >= self._size

    @property
    def fields(self) -> [(int, int)]:
        '''
        Return the list of fields this Ship occupies as a list of (x, y)
        tuples.

        :type: [(int, int)]
        '''
        pos_x, pos_y = self._x, self._y
        result = [(pos_x, pos_y)]

        if self._horizontal:
            deltax = 1 # Go left
            deltay = 0
        else:
            deltax = 0
            deltay = 1 # Go down

        for _ in range(self._size-1):
            pos_x += deltax
            pos_y += deltay

            result.append(
                (pos_x, pos_y)
            )

        return result

    def process_hit(self, x: int, y: int) -> ShotResult:
        '''
        Assert that the given shot was on this Ship, if so, increase the
        hitcounter. Return whether the Ship was missed, hit, or sunk.
        '''

        hit_ship = \
            all([
                self._horizontal,
                y == self._y,
                x >= self._x,
                x < (self._x + self._size)
            ]) or \
            all([
                not self._horizontal,
                x == self._x,
                y >= self._y,
                y < (self._y + self._size)
            ])

        if hit_ship:
            self._hit += 1
            if self.sunk:
                return ShotResult.SUNK
            return ShotResult.HIT

        return ShotResult.MISS

    @property
    def size(self):
        '''
        Return the number of fields the Ship occupies.

        :type: int
        '''
        return self._size

    notational_pattern = r'''
\(
(?P<x>[A-Ja-j])
(?P<y>([1-9]|10)),\s*
(?P<orientation>[hHvV]),\s*
(?P<size>[2-5])
\)
    '''
    notational_regex = re.compile(notational_pattern, re.X)

    @classmethod
    def parse_notation(cls, notation: str):
        '''
        Parse a written representation of a Ship, like `(A1, h, 2)`,
        return a new Ship on this position.

        :param str notation: the notation to parse.
        :raises ValueError: if the passed notation is invalid.
        '''

        match = cls.notational_regex.match(notation)

        if not match:
            raise ValueError("Invalid notation!")

        x = ord(match.group('x').upper()) - ord('A')
        y = int(match.group('y'))         - 1

        horizontal = match.group('orientation').upper() == 'H'

        size = int(match.group('size'))

        return Ship(x, y, horizontal, size)

    def __repr__(self):
        return '<Ship({}, {}, {}, {})>'.format(
            self._x, self._y,
            self._horizontal, self._size
        )

    shot_pattern = r'''
(?P<x>[A-Ja-j])
(?P<y>([1-9]|10))
    '''
    shot_regex = re.compile(shot_pattern, re.X)

    @classmethod
    def parse_shot_notation(cls, notation: str) -> (int, int):
        '''
        Parse a single location on the board in traditional Battleships
        notation (A1-J10), and return a (x, y)-tuple of indexes on the Board.

        :param str notation: the notations to parse.

        :raises ValueError: if the passed notation is invalid.
        '''

        match = cls.shot_regex.match(notation)

        if not match:
            raise ValueError("Invalid notation!")

        x = ord(match.group('x').upper()) - ord('A')
        y = int(match.group('y'))         - 1

        return (x, y)



from setuptools import setup

from pybattleships import VERSION

setup(name='pybattleships',
      version=VERSION,
      description='A Python Battleship framework',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
      ],
      keywords='sticky woestgaaf battleship',
      url='http://github.com/RobinSikkens/pybattleships',
      author='Robin Sikkens & Maarten van den Berg',
      author_email='pybattleships@robinsikkens.nl',
      license='MIT',
      packages=['pybattleships'],
      install_requires=[
      ],
      include_package_data=True,
      zip_safe=False
)
